/* -*-mode: C; style: K&R; c-basic-offset: 4 ; -*- */

/**
 *   List all open top level X windows, and allow them to be killed
 * by their title.
 *
 * Steve
 * ---
 * www.steve.org.uk
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <regex.h>


#include <X11/Xos.h>
#include <X11/Xlib.h>
#include <X11/cursorfont.h>
#include <X11/Xproto.h>
#include <X11/keysymdef.h>
#include <X11/Xmu/WinUtil.h>



/**
 * Prototypes.
 */
int kill_windows (Display *dpy, int screenno, char *pattern, int list);
int catch_window_errors (Display *dpy, XErrorEvent *ev);
int match_regexp( char *pattern, char *title );
void show_usage( int argc, char **argv);
void show_version( int argc, char **argv);


/**
 * Show the version information for this program.
 */
void show_version( int argc, char **argv)
{
    printf( "xclose version %s - compiled on %s\n",
	    VERSION, __DATE__ );
}

/**
 * Show the usage information for this program.
 */
void show_usage( int argc, char **argv)
{

    printf ("xclose - Close XWindows by regular expression.\n");
    printf ("usage: xclose [--list] regexpression\n\n" );
    printf (" --list     List matching windows, don't kill anything.\n" );
    printf (" --help     Shows this usage information.\n" );
    printf (" --version  Shows the version number of xclose.\n\n" );

    /* Show the version. */
    show_version( argc, argv);
    printf ("Report bugs to Steve Kemp <xclose@steve.org.uk>\n" );

}


/**
 * Entry point to the program.
 */
int main (int argc, char * argv[])
{
    char *display = NULL;    /* name of server to contact */
    Display *dpy  = NULL;    /* The display */
    char *pattern = NULL;    /* Regexp of windows to kill */
    int list      = 0;       /* Should we list matching windows. */
    int screenno  = 0;       /* screen number of display */
    int i;
    
    /*
     * First parse any arguments
     */
    for( i = 1; i < argc; i++ )
    {
	if ( strcmp( argv[ i ], "--help" ) == 0 )
	{
	    show_usage( argc, argv );
	    return 0;
	}
	else if ( strcmp( argv[ i ], "--version" ) == 0 )
	{
	    show_version( argc, argv );
	    return 0;
	}
	else if ( strcmp( argv[ i ], "--list" ) == 0 )
	{
	    list = 1;
	}
	else 
	{
	    pattern = argv[ i ];
	}
    }

    dpy = XOpenDisplay (display);
    if (!dpy) 
    {
	fprintf (stderr, "%s:  unable to open display \"%s\"\n",
		 argv[0], XDisplayName (display));
	return( -1 );
    }
    
    screenno = DefaultScreen (dpy);

    kill_windows (dpy, screenno, pattern, list );

    XCloseDisplay (dpy);
    
    return( 0 );
}



/**
 * Handler for any X errors we receive.  We just ignore them
 * as there's nothing we really care about.
 */
int catch_window_errors (Display *dpy, XErrorEvent *ev)
{
    return 0;
}


/**
 * Test a window title against the given pattern.
 */
int match_regexp( char *pattern, char *title )
{
    regex_t compiled;
    int result = 0;

    result = regcomp (&compiled, pattern, REG_EXTENDED );

    /*
     * Generate a sensible error message if the pattern couldn't
     * be compiled.
     */
    if ( result != 0 )
    {
	size_t length = regerror (result, &compiled, NULL, 0);
	char *buffer = (char *)malloc (length);
	(void) regerror (result, &compiled, buffer, length);
	
	fprintf( stderr, 
		 "Failed to compile the pattern '%s' "
		 "into a regular expression:\n%s\n", 
		 pattern, buffer );
	exit( -1 );
    }

    /*
     * See if there's a match.
     */
    result = regexec (&compiled, title, 0, NULL, 0);
    
    regfree( &compiled );

    return( result == 0 );
}



/**
 *  Enumerate all the visible top level windows and kill those
 * that match the given pattern.
 *
 *  If 'list' is non-zero just display the windows.
 */
int kill_windows(Display *dpy, int screenno, char *pattern, int list )
{
    Window root = RootWindow (dpy, screenno);
    Window dummywindow;
    Window *children;
    unsigned int nchildren;
    unsigned int i;
    XWindowAttributes attr;

    XSync (dpy, 0);
    XSetErrorHandler (catch_window_errors);

    nchildren = 0;

    /* Query the root window. */
    XQueryTree (dpy, root, &dummywindow, &dummywindow, &children, &nchildren);
    
    /* Save all valid windows in our list of children */
    for (i = 0; i < nchildren; i++) 
    {
	if ( (XGetWindowAttributes(dpy, children[i], &attr) ) &&
	     (attr.map_state == IsViewable) )
	{
	    children[i] = XmuClientWindow(dpy, children[i]);
	}
	else
	{
	    children[i] = 0;
	}
    }

    /*
     * Iterate over each visible child window.
     */
    for (i = 0; i < nchildren; i++) 
    {
	if ( children[i] )
	{
	    /* Get the title */
	    XTextProperty t;
	    XGetWMName( dpy, children[i], &t );

	    if ( ( list ) && 
		 ( t.value != NULL ) )
	    {
		if ( pattern == NULL )
		{
		    printf("%s\n", t.value );
		}
		else
		{
		    if ( match_regexp( pattern, t.value ) )
		    {
			printf("%s\n", t.value );
		    }
		}
	    }
	  else
	  {
	      if ( ( pattern != NULL ) && 
		   ( t.value != NULL ) )
	      {
		  if ( match_regexp( pattern, t.value ) )
		  {
		      XKillClient (dpy, children[i]);
		  }
	      }
	  }
	}
    }

    /* Free the memory. */
    XFree ((char *)children);

    XSync (dpy, 0);
    XSetErrorHandler (NULL);
    
    return 0;
}
